/* Extract the color information from a Kodak DCS-420 into a RGB format */
/* Dan Christian	5-1-96 */
/* The imager array is set up in uniform blocks like this */
/* G R G R ... G R */
/* B G B G ... B G */
/* G R G R ... G R */
/* B G B G ... B G */
/* ... */
/* G R G R ... G R */
/* B G B G ... B G */
/* the raw data will be RGB 1536/2 x 1024/2 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>			/* for lseek */
#include <fcntl.h>			/* for open */
#include <sys/uio.h>			/* for read, write */

#define WIDTH 1536			/* image array width (columns) */
#define HEIGHT 1024			/* image array height (lines) */
#define START 75776			/* offset into DCS420 TIF file */
#define MARGIN 6			/* pixels to ignore around edges */

#ifndef MIN
#define MIN(a,b)	((a<b) ? a : b)
#endif

void usage (void)
{
    fprintf (stderr, "Usage: dcs420ctoraw [-maxrgb R G B] file\n");
    exit (1);
}

int
main (int argc, char **argv)
{
    int		fd;			/* input file descriptor */
    int		outFd;			/* output file descriptor */
    int 	r;			/* general response variable */
    int 	line;			/* line counter */
    unsigned int redMax=0, greenMax=0, blueMax=0; /* track maximums */
    int		colorCorrect = 1;	/* do color correction */
    int		autoBalance = 1;	/* find Maxes for balance */
    unsigned char	image[WIDTH*2];		/* two lines of the image */
    unsigned char *green = image+0;		/* pixel pointers */
    unsigned char *red = image+1;
    unsigned char *blue = image+WIDTH+0;
    unsigned char *green2 = image+WIDTH+1;	/* second green */
    
    --argc, ++argv;			/* skip past prog name */
    while (argc && ('-' == *argv[0])) {
	if (!strcmp (argv[0], "-maxrgb")) { /* set redMax, greenMax, blueMax */
	    if (argc < 3) usage ();
	    sscanf (argv[1], "%d", &redMax);
	    sscanf (argv[2], "%d", &greenMax);
	    sscanf (argv[3], "%d", &blueMax);
	    argc -= 3;
	    argv += 3;
	    autoBalance = 0;
	    fprintf (stderr, "Setting red max %d green max %d blue max %d\n",
		     redMax, greenMax, blueMax);
	}
	--argc, ++argv;
    }

    if (argc < 1) usage ();		/* enough arguments? */
    fd = open (argv[0], O_RDONLY, 0);
    if (fd < 0) { perror (argv[0]); exit (2); }	/* print and die */

    outFd = fileno (stdout);		/* output file descriptor */
    if (outFd < 0) { perror ("fileno"); exit (2); } /* print and die */

    r = lseek (fd, START, SEEK_SET); /* skip to the image array */
    if (r < 0) { perror ("lSeek"); exit (3); }	/* print and die */

    if (autoBalance) {
	for (line = 0; line < HEIGHT; line += 2) { /* find white ballance */
	    int col;
	
	    r = read (fd, image, WIDTH*2);
	    if (r < 0) { perror ("read"); exit (4); } /* print and die */
	    green = image+0;		/* pixel pointers */
	    red = image+1;
	    blue = image+WIDTH+0;
	    green2 = image+WIDTH+1;	/* second green */
	
	    for (col = 0; col < WIDTH/2; ++col) {
		if (*red > redMax) redMax = *red;
		if (((int)*green+*green2)/2 > greenMax)
		    greenMax = ((int)*green+*green2)/2;
		if (*blue > blueMax) blueMax = *blue;
		red += 2;		/* increment to next pixel block */
		green += 2;
		blue += 2;
		green2 += 2;
	    }
	}
	fprintf (stderr, "Red max: %d Green max: %d Blue max: %d\n",
		 redMax, greenMax, blueMax);
    }

    if (colorCorrect) {
	if ((255 == redMax) && (255 == greenMax) && (255 == blueMax)) {
	    colorCorrect = 0;		/* no correction to do */
	    fprintf (stderr, "No color correction possible\n");
	}
    }
    r = lseek (fd, START, SEEK_SET); /* skip to the image array */
    if (r < 0) { perror ("lSeek"); exit (3); }	/* print and die */

    for (line = 0; line < HEIGHT; line += 2) {
	unsigned char	imageOut[3*WIDTH/2];	/* output array */
	unsigned char	*out = imageOut;
	int col;
	
	r = read (fd, image, WIDTH*2);
	if (r < 0) { perror ("read"); exit (4); } /* print and die */
	green = image+0;		/* pixel pointers */
	red = image+1;
	blue = image+WIDTH+0;
	green2 = image+WIDTH+1;	/* second green */
	
	for (col = 0; col < WIDTH/2; ++col) {
	    unsigned int	p;
	    p = *red;
	    if (colorCorrect) p = (p * 256) / redMax; /* color correct */
	    if (p > 255) p = 255;
	    *out++ = p;

	    p = (unsigned int)*green + *green2;	/* average the two greens */
	    if (colorCorrect) p = (p * 128) / greenMax; /* color correct */
	    if (p > 255) p = 255;
	    *out++ = p;

	    p = *blue;
	    if (colorCorrect) p = (p * 256) / blueMax; /* color correct */
	    if (p > 255) p = 255;
	    *out++ = p;

	    red += 2;			/* increment to next pixel block */
	    green += 2;
	    blue += 2;
	    green2 += 2;
	}
	r = write (outFd, imageOut, sizeof (imageOut));
	if (r < 0) { perror ("write"); exit (4); } /* print and die */
    }
    return 0;
}
