/* Find first/all files in path matching a simple glob pattern */
/* Dan Christian	4-18-96 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define MAXENTRY	128
int	debugLevel = 0;

void
usage (void)
{
    fprintf (stderr, "Usage: \tfindpath [options] [path]* [pattern]*\n");
    fprintf (stderr,
	     "Findpath finds files matching pattern(s) in a path\n");
    fprintf (stderr, "\t-f \tjust return the first file found\n");
    fprintf (stderr, "\t-dCHAR \tadd CHAR as a path separator\n");
    fprintf (stderr, "\t-LDIR \tadd DIR to the end of the search path\n");
    fprintf (stderr, "\t-IDIR \tadd DIR to the end of the search path\n");
    fprintf (stderr, "\t-lLIB \tsearch for libLIB.a and libLIB.so\n");
    fprintf (stderr, "\t-p \tPrint path Purged of invalid directories\n");
    fprintf (stderr, "\t-v \tPrint extra debugging information\n");
    fprintf (stderr,
	     "The wildcards * and ? are allowed in the pattern.\n");
    fprintf (stderr,
	     "Search pathes use colons or spaces as separators \n");
    fprintf (stderr,
	     "separated list of directories (e.g. /usr/bin:/bin).\n");
    fprintf (stderr, "Usage: \tfindpath pattern\n");
    fprintf (stderr, "Findpath finds files matching pattern in $PATH\n");
    fprintf (stderr,
	     "Example: \"findpath $PATH '*csh'\" will find variants on csh.\n");
    fprintf (stderr,
	     "Example: \"findpath -L/usr/lib/X11 /lib:/usr/lib '-lX*'\" will find X libraries.\n");
    fprintf (stderr,
	     "Example: \"findpath csh\" finds locations of csh\n");
    exit (1);
}

/* Recursive string matching supporting * and ? wildcards */
int
SubMatch (pattern, string)
  char	*pattern, *string;
{
    char	*src, *dst;

    for (src = pattern, dst= string; *src || *dst; ) {
	if (*src == '*') {	/* match zero or more characters */
	    ++src;		/* next real character */
	    for (;;) {
		while (*dst && (*dst != *src)) ++dst;
		if (!*dst) break; /* end of string */
		if (SubMatch (src, dst)) break; /* match */
		else ++dst;	/* remainder did not fully match */
	    }
	} else if (*src == '?') { /* match any one character */
	    ++src; ++dst;
	} else {
	    if (*src == *dst) {
		++src;
		++dst;
	    } else {
		if (debugLevel >= 3)
		  printf ("Mismatch of %s %s at <%s> <%s>\n",
			  pattern, string, src, dst);
		break;
	    }
	}
    }
    if (!*src && !*dst) return (1);	/* matched */
    return (0);				/* failed */
}

/* Search through a directory for first/all entries matching pattern */
int
CheckDir (char *pattern, char *dir, int matchAll)
{
    DIR		*dirList;
    struct dirent	*dirEnt;
#ifdef JUST_FILES
    struct stat		statInfo;
#endif
    int	cnt=0;

    if (!(dirList = opendir (dir))) {
	char	berror[256];		/* error printing buffer */
	sprintf (berror, "opendir: %s", dir);
	perror (berror);
	return (-1);
    }
    for (; dirEnt = readdir (dirList);) {
	if ((!strcmp (dirEnt->d_name, ".")) ||
	    (!strcmp (dirEnt->d_name, "..")) )
	  continue;
	
#ifdef JUST_FILES
	if (stat (buff, &statInfo) < 0) { /* get first file type */
	    char	berror[256];	/* error printing buffer */
	    sprintf (berror, "stat: %s", buff);
	    perror (berror);
	    continue;
	}
	if (S_ISDIR (statInfo.st_mode)) {
	    continue;
	}
#endif
	if (SubMatch (pattern, dirEnt->d_name)) {
	    printf ("%s/%s\n", dir, dirEnt->d_name);
	    ++cnt;
	    if (!matchAll) break;
	}
    }
    closedir (dirList);
    return cnt;
}

/* Search a series of directories for first/all entries matching filePat */
int
CheckDirList (char *filePat,		/* file to match */
	      char **dirs,		/* series of directories */
	      int dirCnt,		/* number of directories */
	      int showAll)		/* whether to stop at the first match*/
{
    int	cnt=0, i, r;

    if (dirCnt <= 0) return 0;
    for (i = 0; i < dirCnt; ++i) {
	if (debugLevel >= 2) {		/* show the path for debug */
	    printf ("Checking: %s\n", dirs[i]);
	}
	r = CheckDir (filePat, dirs[i], showAll);
	if (r > 0) {
	    cnt += r;
	    if (!showAll) return cnt;	/* got first one, done */
	}
    }
    return cnt;				/* return number of matches */
}

/* Search a series of directories and print valid ones */
int
PurgeDirList (char **dirs,		/* series of directories */
	      int dirCnt)		/* number of directories */
{
    int	cnt=0, i, r;

    if (dirCnt <= 0) return 0;
    for (i = 0; i < dirCnt; ++i) {
	DIR	*dirList;
	if (debugLevel >= 2) {		/* show the path for debug */
	    printf ("Checking: %s\n", dirs[i]);
	}
	if (dirList = opendir (dirs[i])) {
	    if (cnt) fputs (":", stdout); /* insert separator */
	    fputs (dirs[i], stdout);
	    closedir (dirList);
	    ++cnt;
	}
    }
    if (cnt) puts ("\n");
    return cnt;				/* return number of matches */
}

char pathSep[64] = ": ";		/* possible path separators */

/* Given a path, break it a separators and stick the fields in dirs */
/* return the new directory count */
int
PathAdd (char *path, char **dirs, int dirCnt)
{
    char *cp, *start;
    for (cp = start = path; ; ++cp) { /* split path into directories */
	if ((*cp == 0) || strchr (pathSep, *cp)) {
	    if (cp != start) {		/* non empty field */
		dirs[dirCnt++] = start;
	    }
	    if (!*cp) break;	/* done */
	    *cp = 0;		/* terminate string */
	    start = cp+1;
	}
    }
    return dirCnt;
}

main (int argc, char **argv)
{
    char	*dirs[MAXENTRY];
    int	dirCnt=0;
    int	showAll = 1;
    int	found = 0;
    char	*cp, *start;
    int	i, r;

    --argc, ++argv;

    if (!argc) usage();			/* no arguments, print help */


    while (argc) {			/* parse switches, pathes, patterns */
	if (**argv == '-') {		/* a switch */
	    cp = *argv;
	    ++cp;
	    while (*cp) {
		switch (*cp) {

		case 'a':	++showAll; break;

		case 'f':	showAll = 0; break;

		case 'd': {		/* add separator */
		    char *ep;
		    ++cp;
		    if (!*cp) usage();
		    for (ep = pathSep; *ep; ++ep); /* go to the end */
		    *ep++ = *cp;	/* add separator to the end */
		    *ep++ = 0;
		}

		case 'v':	++debugLevel; break;

		case 'I':		/* -Idirectory style */
		case 'L':		/* -Ldirectory style */
		    ++cp;
		    if (debugLevel)
			fprintf (stderr, "Adding path <%s>\n", cp);
		    dirs[dirCnt++] = cp; /* get directory past -L */
		    goto nextArgument;
		    break;

		case 'l': {		/* -llib style */
		    char	fullName[256];
		    sprintf (fullName, "lib%s.so", cp+1);
		    if (debugLevel)
			fprintf (stderr, "Searching on <%s>\n", fullName);
		    r = CheckDirList (fullName, dirs, dirCnt, showAll);
		    found += r;
		    sprintf (fullName, "lib%s.a", cp+1);
		    if (debugLevel)
			fprintf (stderr, "Searching on <%s>\n", fullName);
		    r = CheckDirList (fullName, dirs, dirCnt, showAll);
		    found += r;
		    goto nextArgument;
		}

		case 'p': {
		    if (!dirCnt) {		/* default path */
			cp = getenv ("PATH");
			if (!cp) {
			    fprintf (stderr, "Could not get $PATH!\n");
			    exit (-1);
			}
			if (debugLevel) 
			    fprintf (stderr, "Using defaultpath <%s>\n", cp);
			dirCnt = PathAdd (cp, dirs, dirCnt);
		    }
		    found += PurgeDirList (dirs, dirCnt);
		    goto nextArgument;
		}

		case 'h':	usage(); /* which exits */

		default:
		    printf ("Ignoring argument %c, Use findpath -h for help\n",
			    *cp);
		}
		++cp;
	    }
	} else {			/* path */
	    for (cp = *argv; *cp; ++cp) {
		if (strchr (pathSep, *cp)) { /* its a path */
		    if (debugLevel)
			fprintf (stderr, "Parsing path <%s>\n", *argv);
		    dirCnt = PathAdd (*argv, dirs, dirCnt);
		    goto nextArgument;
		}
	    }

	    if (!dirCnt) {		/* default path */
		cp = getenv ("PATH");
		if (!cp) {
		    fprintf (stderr, "Could not get $PATH!\n");
		    exit (-1);
		}
		if (debugLevel) 
		    fprintf (stderr, "Using defaultpath <%s>\n", cp);
		dirCnt = PathAdd (cp, dirs, dirCnt);
		if (CheckDirList (*argv, dirs, dirCnt, 1) > 0) {
		    return 0;
		} else {
		    return -1;
		}
	    }

	    if (!*cp) {			/* its a pattern to search for */
		if (debugLevel)
		    fprintf (stderr, "Searching on <%s>\n", *argv);
		r = CheckDirList (*argv, dirs, dirCnt, showAll);
		if ((r > 0) && !showAll) return 0; /* found, return success */
		found += r;
	    }
	}

    nextArgument:
	--argc, ++argv;
    }
    if (found) return 0;
    return -1;
}
