/*	Ushell  -> a simple shell by Dan Christian
	Ushell allows commands to be executed with arguments and allows
	the output of one command to be piped into a subsequent command.
	Ushell supports advanced editing on h19 and vt100 terminal 
	(and variants).

Sample command line:
	ls -la .. | fgrep x
Would show all the executable files (plus a few others) in the directory
above the current one.

Editing available:
	^f	move forward a character
	^b	move backward a character
	^a	move to the beginning of the line
	^e	move to the end of the line
	^d	delete the character under cursor
	<backspace>,<delete>	delete character to the left of cursor
	^u	clear entire line
	^w	erase word to left of cursor
	^r	redisplay the line
	^l	clear screen
	^p	substitute a command from the command ring and rotate ring up
	^n	substitute a command from the command ring and rotate ring down

Currently, editing is supported for h19 (also z19, vt52, etc) terminals and
vt100 terminals (also vt102, and xterm).  Other terminal types remain in 
the standard tty input mode.
*/

#include <stdio.h>
#include <string.h>
#include <sys/file.h>
#include <edring.h>

#define R_SZ 8
#define B_SZ 80

/* These macros allow me to automatically dynamically allocate an array
 in chunks of ARRAY_BLK.  It guarentees that the array will have room
 for at least n+1 items. start_array must be used before build_array.
	a is some array. n is the highest index to be used.
	t is a typecast for the malloc.  
	NOTE: there must be sufficient memory: malloc errors are ignored.
*/


#define ARRAY_BLK 4		/* must be a power of 2 */
#define start_array(a,t) (a)= (t) malloc (sizeof (t) * ARRAY_BLK)
#define build_array(a,n,t) ((n) >= ((n)+(ARRAY_BLK -1) & ~(ARRAY_BLK -1))) ? \
  (a) = (t) realloc ((a), sizeof (t) * ((n)+ARRAY_BLK & ~(ARRAY_BLK -1))) : (a)

static int debug;		/* a debuging level flag */

char *skip_over(), *skip_to();

/*--------------------------- parse_args ---------------*/
/*
  parse_args breaks a string, str, into substrings at whitespace.  These
  substrings are placed in an array of strings.  The last entry in the
  array is a null pointer.

  parse_args finishes when it finds a '|' or the end of the string.
  parse_args returns the string following the '|' or a pointer to
  a empty string.
*/
char *parse_args (str, alist)	/* break str into seperate arguments */
  char *str, ***alist;		/* alist is a pointer to an array of strings */
{
    char *begin, *end;
    int cnt=0;
    int len;

    end = str;
    start_array ((*alist), char**);
    do {
	begin = skip_over (end, " \t"); /* find beginning of a word */
	if (*begin == 0) {
	    end = begin;
	    break;	/* nothing left */
	}
	if (*begin != '|') {
	    end = skip_to (begin, " \t|");	/* find end of word */
				/* find string length */
	    len = (int)((unsigned int)end - (unsigned int)begin);

	    build_array ((*alist), cnt, char **);
	    (*alist)[cnt] = (char *)malloc (len+1);
	    strncpy ((*alist)[cnt], begin, len);
	    (*alist) [cnt][len] = 0;	/* null terminate strings */

	    if (debug > 3) puts ((*alist)[cnt]);
	    cnt++;
	} else {
	    end = skip_over (begin, " \t|");
	    build_array ((*alist), cnt, char **);
	    (*alist) [cnt] = 0;
	    return (end);	/* return pointer to rest of string */
	}
	
    } while (*end);

    build_array ((*alist), cnt, char **);
    (*alist) [cnt] = 0;
    return (end);
}

/*------------------------- skip_to -----------------*/
/*
  skip_to returns a pointer to the first occurance of a character in
  the string, charset, in the string, string.  If there are no occurances
  of the characters in charset in string, a pointer to the null at
  the end of the string is returned.
  */
char *skip_to (string, charset)	/* skipto first char in charset */
    char  *string,
          *charset;
{
    char *p, *i;

    for (p = string ; *p; p++)	/* scan string */
      for (i = charset; *i; i++) /* scan charset */
	if (*i == *p) return (p); /* exit on first occurance */

    return (p);
}

/*------------------------- skip_over ---------------*/
/*
  skip_over is the complement to skip_to.  It returns a pointer to
  the first character not in charset.
  */
char *skip_over (string, charset) /* skip while chars are in charset */
    char  *string,
	  *charset;
{
    char *p, *i;

    for (p = string ; *p; p++) { /* scan strin */
	for (i=charset; *i; i++) /* scan charset */
	  if (*i == *p) break;
	if (*i == 0) break;	/* exit if not in charset */
    }
    return (p);
}

print_args (argv)		/* print arguments for debugging */
  char **argv;
{
    while (*argv)
      fprintf (stderr, "%s\n", *(argv++));
}

/*------------------------- run_child ---------------*/
/*
  run_child takes an array of arguments, a file to be used as input
  to the process, and a flag telling it to create a pipe for the output
  of the process to be run.

  If fin <0, then stdin is used for input.

  run_child returns the output filedescriptor from any pipes that it
  created.
  */

int run_child (arg, fin, pipeit) /* returns output file  */
  char **arg;
  int fin;			/* if >= 0, use for inptu */
  int pipeit;			/* if !=0 make a pipe and return output */
{
    int pio[2];

    if (fin >= 0)
      if ( dup2 (fin, 0) < 0) {	/* duplicate stdin */
	  fputs ("Unable to set input.\n", stderr);
      } else {
	  if (debug > 1) fputs ("Opened input.\n", stderr);
      }
    close (fin);

    if  (pipeit == 0) { /* easy case: just run it */
	if (debug > 2) fputs ("Hi granddad!\n", stderr);
	execvp (arg[0], arg);
	fprintf (stderr, "Could not execute: %s\n", arg[0]);
	exit (-1);
    }

    
    if (pipe (pio)) {		/* open pipe for output */
	fprintf (stderr, "Error opening pipe.\n");
	exit (-1);
    }
    
    if (fork () == 0) {		/* supply side of pipe */
	if (debug > 2) fprintf (stderr, "Grandpapa!\n");
	
	if (dup2 (pio[1], 1) < 0) { /* duplicate stdout */
	    fputs ("Unable to set output.\n", stderr); /* duplicate stdin */
	} else {
	    if (debug > 1) fputs ("Opened output.\n", stderr);
	}
	close (pio[1]);
	
	execvp (arg[0],arg);
	fprintf (stderr, "Could not execute: %s\n", arg[0]);
	exit (-1);
    } else {
	close (pio[1]);
	return (pio[0]);	/* return output of pipe */
    }
}

/*-------------------------- main ----------------------- */
/*
  main reads a line from stdin and then forks off a child to execute
  everything specified on the command line.  The prompt will return
  immeadiately on blank lines.  ushell returns when "exit" is typed
  or EOF is encountered.
  */
main (argc, argv)
  int argc;
  char **argv;
{
    char ring[R_SZ][B_SZ], *comm; /* buffer ring for rling */
    int flags = 0, i, j, editing = 0;
    
    char **arg, *cp;
    int pio[2];
    int inpipe = -1;

    debug = argc;		/* hack to set debugging level */
    
    comm = &ring[0][0];
				/* set up terminal for editing */
    if (strcmp (getenv ("TERM"), "wm")) {
	editing = 1;
    } else {			/* wm won't allow editing */
	fputs ("Warning: only simple editing enabled.\n", stderr);
    }

    while (1) {
	if (editing) {
	    comm = edring (stdin, stdout, ":-) ", /* get a line with editing */
			  comm, B_SZ, ring, R_SZ, flags);
	} else {
	    fputs (":-) ", stdout);
	    gets (comm);
	}

	if (comm[0] == 0) {	/* Empty input line */
	    if (debug > 2) fprintf (stderr, "louder please.\n");
	    continue;
	}
	
	if ((strncmp (comm, "exit", 4) == 0) || feof (stdin))
	  return; /* exit on "exit" */
	
	if (fork () != 0) {
	    if (wait (0) < 0) {		/* parent process*/
		fprintf (stderr, "Error in child process.\n");
	    }
/* execute processes */
	} else {		/* run processes */
	    cp = comm;		/* child */
	    if (debug > 2) fprintf (stderr, "Hi dad\n");
	    
	    while (*cp) {	/* while there is string left */
		if (debug > 2) fprintf (stderr, "Parsing: %s\n", cp);
		cp  = parse_args (cp, &arg);
		if (debug > 2) print_args (arg);
		if (debug > 2) fprintf (stderr, "Will parse: %s\n", cp);

		inpipe  = run_child (arg, inpipe, *cp);
		/* as long as *cp != 0; pipe.  inpipe is the input
		   from the last pipe (or -1 if none). */
	    }
	}
    }
}
